from __future__ import absolute_import

import os
import gensim
import pandas as pd

from topik.intermediaries.digested_document_collection import DigestedDocumentCollection
from topik.intermediaries.raw_data import load_persisted_corpus
from .model_base import TopicModelBase, register_model

# Doctest imports
from topik.readers import read_input
from topik.tests import test_data_path


@register_model
class LDA(TopicModelBase):
    """A high-level interface for an LDA (Latent Dirichlet Allocation) model.


    Parameters
    ----------
    corpus_input : CorpusBase-derived object
        object fulfilling basic Corpus interface (preprocessed, tokenized text).
        see topik.intermediaries.tokenized_corpus for more info.
    ntopics : int
        Number of topics to model
    load_filename : None or str
        If not None, this (JSON) file is read to determine parameters of the model persisted to disk.
    binary_filename : None or str
        If not None, this file is loaded by Gensim to bring a disk-persisted model back into memory.


    Attributes
    ----------
    corpus : CorpusBase-derived object, tokenized
    model : Gensim LdaModel instance


    Examples
    --------
    >>> raw_data = read_input('{}/test_data_json_stream.json'.format(test_data_path), "abstract")
    >>> processed_data = raw_data.tokenize()  # preprocess returns a DigestedDocumentCollection
    >>> model = LDA(processed_data, ntopics=3)

    """
    def __init__(self, corpus_input=None, ntopics=10, load_filename=None, binary_filename=None, **kwargs):
        if corpus_input is not None:
            self._model = gensim.models.LdaModel(list(iter(corpus_input)), num_topics=ntopics,
                                                id2word=corpus_input.get_id2word_dict(), **kwargs)
            self._corpus = corpus_input
        elif load_filename is not None and binary_filename is not None:
            self._model = gensim.models.LdaModel.load(binary_filename)
            self._corpus = DigestedDocumentCollection(load_persisted_corpus(load_filename))

    def save(self, filename):
        self._model.save(self.get_model_name_with_parameters())
        saved_data = {"load_filename": filename, "binary_filename": self.get_model_name_with_parameters()}
        return super(LDA, self).save(filename, saved_data)

    def get_top_words(self, topn):
        top_words = [self._model.show_topic(topicno, topn) for topicno in range(self._model.num_topics)]
        return top_words

    def get_model_name_with_parameters(self):
        return "LDA_{}_topics{}".format(self._model.num_topics, self._corpus.filter_string)

    def _get_topic_term_dists(self):
        term_topic_df = pd.DataFrame([
                pd.DataFrame.from_records(self._model.show_topic(topic_no, None),
                                         columns=['topic' + str(topic_no) + 'dist', 'token'],
                                         index='token')['topic' + str(topic_no) + 'dist']
                for topic_no in range(self._model.num_topics)]).T
        term_topic_df['term_id'] = pd.Series(dict(self._corpus._dict.token2id.items()))
        term_topic_df = term_topic_df.set_index('term_id')
        return term_topic_df

    def _get_doc_topic_dists(self):
        id_index, bow_corpus = zip(*[(id, self._corpus._dict.doc2bow(doc_tokens))
                              for id, doc_tokens in self._corpus._corpus])

        doc_topic = list(self._model[bow_corpus])

        for i, doc in enumerate(doc_topic):
            for j, topic in enumerate(doc):
                doc_topic[i][j] = doc_topic[i][j][1]

        doc_topic_df = pd.DataFrame(doc_topic, index=id_index)
        doc_topic_df.columns = ['topic'+str(i)+'dist' for i in range(
                                                doc_topic_df.shape[1])]
        doc_topic_df.index.name = 'doc_id'
        return doc_topic_df



import unittest
import os
import time

import elasticsearch

from topik.readers import read_input
from topik.intermediaries.raw_data import load_persisted_corpus, ElasticSearchCorpus
from topik.tests import test_data_path


INDEX = "topik_unittest"
SAVE_FILENAME = "test_save"

class BaseCorpus(object):
    test_raw_data = None

    def test_year_filtering(self):
        result_list = list(self.test_raw_data.get_date_filtered_data(start=1975,
                                                                     end=1999,
                                                                     field="year"))
        self.assertEqual(len(result_list), 25)
        self.assertTrue(-1611117933394825767 in [int(item[0]) for item in
                        result_list])

    def test_generator_without_id(self):
        data = list(self.test_raw_data.get_generator_without_id())
        self.assertEqual(len(data), 100)
        self.assertFalse(data[0] == data[1])

    def test_save_file(self):
        self.test_raw_data.save(SAVE_FILENAME)
        self.assertTrue(os.path.exists(SAVE_FILENAME))

    def test_load_file(self):
        self.test_raw_data.save(SAVE_FILENAME)
        test_raw_data = load_persisted_corpus(SAVE_FILENAME)
        self.assertEqual(len(test_raw_data), 100)

    def test_has_filter_string(self):
        self.assertEqual(self.test_raw_data.filter_string, "")
        date_data = self.test_raw_data.get_date_filtered_data(1970, 2000)
        self.assertNotEqual(date_data.filter_string, "")


class TestDictionaryCorpus(unittest.TestCase, BaseCorpus):
    def setUp(self):
        self.test_raw_data = read_input('{}/test_data_json_stream.json'.format(
            test_data_path), content_field="abstract")


class TestElasticSearchCorpus(unittest.TestCase, BaseCorpus):
    def setUp(self):
        self.test_raw_data = read_input('{}/test_data_json_stream.json'.format(
            test_data_path), content_field="abstract",
            output_type=ElasticSearchCorpus.class_key(),
            output_args={'source': 'localhost',
                         'index': INDEX},
            synchronous_wait=30)

    def tearDown(self):
        instance = elasticsearch.Elasticsearch("localhost")
        instance.indices.delete(INDEX)
        if instance.indices.exists("{}_year_alias_date".format(INDEX)):
            instance.indices.delete("{}_year_alias_date".format(INDEX))
        time.sleep(1)



import os
import unittest

from topik.readers import read_input
from topik.tests import test_data_path


class TestCorpusBOW(unittest.TestCase):
    def setUp(self):
        self.dictionary_values_simple_test_data_1 = [
                'bending', 'sci', 'forget', 'messi', 'skip',
                'hands', 'focus', 'comply', 'colors', 'planning']

        self.dictionary_values_simple_test_data_json_stream = [
            u'limited', u'consolidated', u'magnetic', u'comparatively',
            u'powders', u'waspaloy', u'tensile', u'assembled', u'relationships',
            u'sfft']

        self.corpus_bow_head_2_simple_test_data_1 = [
            (0, 1), (1, 1), (2, 1), (3, 1), (4, 1), (5, 1),
            (6, 1), (7, 1), (8, 1), (9, 1), (10, 1)]

        self.corpus_bow_head_2_simple_test_data_json_stream = [
            (0, 1), (1, 1), (2, 1), (3, 1), (4, 1), (5, 1), (6, 1), (7, 1),
            (8, 1), (9, 2), (10, 1), (11, 1), (12, 1), (13, 1), (14, 2),
            (15, 2), (16, 1), (17, 1), (18,1), (19, 2), (20, 2), (21, 2),
                (22, 1), (23, 1), (24, 1), (25, 1), (26, 1), (27, 1), (28, 1),
                (29, 1), (30, 1), (31, 1), (32, 1), (33, 1), (34, 1), (35, 1),
                (36, 1), (37, 1), (38, 1), (39, 1), (40, 1), (41, 1), (42, 1)]

        raw_data = read_input(os.path.join(test_data_path,
                                           'test_data_json_stream.json'),
                                   content_field="abstract",
                                   output_type="dictionary")
        self.processed_data = raw_data.tokenize(min_length=1)

    def test_corpus_bow_content(self):
        self.assertEqual(self.processed_data._dict.values()[:10],
                        self.dictionary_values_simple_test_data_json_stream)

    def test_corpus_word_counts(self):
        self.assertEqual(next(iter(self.processed_data)),
                         self.corpus_bow_head_2_simple_test_data_json_stream)

if __name__ == '__main__':
    unittest.main()



"""This file handles the storage of data from loading and analysis.

More accurately, the files written and read from this file describe how to read/write actual data, such that the
actual format of any data need not be tightly defined.
"""

import json


class Persistor(object):
    _storage = {"corpus": None, "models": {}}

    def __init__(self, filename=None):
        if filename:
            self.load_data(filename)

    def store_corpus(self, data_dict):
        self._storage["corpus"] = data_dict

    def store_model(self, model_id, model_dict):
        self._storage["models"][model_id] = model_dict

    def get_corpus_dict(self):
        return self._storage["corpus"]

    def get_model_details(self, model_id):
        return self._storage["models"].get(model_id, None)

    def list_available_models(self):
        return self._storage["models"].keys()

    def persist_data(self, filename):
        with open(filename, "w") as f:
            json.dump(self._storage, f)

    def load_data(self, filename):
        with open(filename, "r") as f:
            self._storage = json.load(f)



import os
import unittest

from topik.readers import read_input
from topik.tokenizers import tokenizer_methods, collect_entities, collect_bigrams_and_trigrams

# sample data files are located in the same folder
module_path = os.path.dirname(__file__)


class TestTokenizers(unittest.TestCase):
    def setUp(self):
        self.solution_simple_tokenizer_test_data_json_stream = [
            u'transition', u'metal', u'oxides', u'considered',
            u'generation', u'materials', u'field', u'electronics',
            u'advanced', u'catalysts', u'tantalum', u'v', u'oxide',
            u'reports', u'synthesis', u'material', u'nanometer', u'size',
            u'unusual', u'properties', u'work', u'present', u'synthesis',
            u'ta', u'o', u'nanorods', u'sol', u'gel', u'method', u'dna',
            u'structure', u'directing', u'agent', u'size', u'nanorods',
            u'order', u'nm', u'diameter', u'microns', u'length', u'easy',
            u'method', u'useful', u'preparation', u'nanomaterials', u'electronics',
            u'biomedical', u'applications', u'catalysts']

        self.solution_collocations_tokenizer_test_data_json_stream = [
            u'transition_metal', u'oxides', u'considered', u'generation',
            u'materials', u'field', u'electronics', u'advanced', u'catalysts',
            u'tantalum', u'v_oxide', u'reports', u'synthesis_material',
            u'nanometer_size', u'unusual', u'properties', u'work_present',
            u'synthesis', u'ta', u'o', u'nanorods', u'sol', u'gel', u'method',
            u'dna', u'structure', u'directing', u'agent', u'size', u'nanorods',
            u'order', u'nm_diameter', u'microns', u'length', u'easy', u'method',
            u'useful', u'preparation', u'nanomaterials', u'electronics', u'biomedical',
            u'applications', u'catalysts']

        self.solution_entities_tokenizer_test_data_json_stream = [
            u'transition', u'metal_oxides', u'generation_materials',
            u'tantalum', u'oxide', u'nanometer_size', u'unusual_properties',
            u'ta_o', u'sol_gel_method', u'dna', u'easy_method',
            u'biomedical_applications']

        self.solution_mixed_tokenizer_test_data_json_stream = [
            u'transition', u'metal', u'oxides', u'generation', u'materials',
            u'tantalum', u'oxide', u'nanometer', u'size', u'unusual', u'properties',
            u'ta_o', u'sol', u'gel', u'method', u'dna', u'easy', u'method',
            u'biomedical', u'applications']

        self.data_json_stream_path = os.path.join(module_path, 'data/test_data_json_stream.json')
        self.data_large_json_path = os.path.join(module_path, 'data/test_data_large_json.json')
        assert os.path.exists(self.data_json_stream_path)
        assert os.path.exists(self.data_large_json_path)

    def test_simple_tokenizer(self):
        raw_data = read_input(
                source=self.data_json_stream_path,
                content_field="abstract",
                output_type="dictionary")
        _, text = next(iter(raw_data))
        doc_tokens = tokenizer_methods["simple"](text, min_length=1)
        self.assertEqual(doc_tokens, self.solution_simple_tokenizer_test_data_json_stream)

    def test_collocations_tokenizer(self):
        raw_data = read_input(
                source=self.data_json_stream_path,
                content_field="abstract",
                output_type="dictionary")
        patterns = collect_bigrams_and_trigrams(raw_data, min_bigram_freq=2, min_trigram_freq=2)
        _, text = next(iter(raw_data))
        doc_tokens = tokenizer_methods["collocation"](text, patterns=patterns)
        self.assertEqual(doc_tokens, self.solution_collocations_tokenizer_test_data_json_stream)

    def test_entities_tokenizer_json_stream(self):
        raw_data = read_input(
                source=self.data_json_stream_path,
                content_field="abstract",
                output_type="dictionary")
        entities = collect_entities(raw_data, freq_min=1)
        _, text = next(iter(raw_data))
        doc_tokens = tokenizer_methods["entities"](text, entities)
        self.assertEqual(doc_tokens, self.solution_entities_tokenizer_test_data_json_stream)

    def test_mixed_tokenizer(self):
        raw_data = read_input(
                source=self.data_json_stream_path,
                content_field="abstract",
                output_type="dictionary")
        entities = collect_entities(raw_data)
        id, text = next(iter(raw_data))
        doc_tokens = tokenizer_methods["mixed"](text, entities)
        self.assertEqual(doc_tokens, self.solution_mixed_tokenizer_test_data_json_stream)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function


import os
import logging

from topik.intermediaries.raw_data import registered_outputs, DictionaryCorpus

# imports used only for doctests
from topik.tests import test_data_path

logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                    level=logging.INFO)


def _iter_document_json_stream(filename, json_prefix=None):
    """Iterate over a json stream of items and get the field that contains the text to process and tokenize.

    Parameters
    ----------
    filename : str
        The filename of the json stream.

    Examples
    --------
    >>> documents = _iter_document_json_stream(
    ... '{}/test_data_json_stream.json'.format(test_data_path))
    >>> next(documents) == {
    ... u'doi': u'http://dx.doi.org/10.1557/PROC-879-Z3.3',
    ... u'title': u'Sol Gel Preparation of Ta2O5 Nanorods Using DNA as Structure Directing Agent',
    ... u'url': u'http://journals.cambridge.org/action/displayAbstract?fromPage=online&aid=8081671&fulltextType=RA&fileId=S1946427400119281.html',
    ... u'abstract': u'Transition metal oxides are being considered as the next generation materials in field such as electronics and advanced catalysts; between them is Tantalum (V) Oxide; however, there are few reports for the synthesis of this material at the nanometer size which could have unusual properties. Hence, in this work we present the synthesis of Ta2O5 nanorods by sol gel method using DNA as structure directing agent, the size of the nanorods was of the order of 40 to 100 nm in diameter and several microns in length; this easy method can be useful in the preparation of nanomaterials for electronics, biomedical applications as well as catalysts.',
    ... u'filepath': u'abstracts/879/http%3A%2F%2Fjournals.cambridge.org%2Faction%2FdisplayAbstract%3FfromPage%3Donline%26aid%3D8081671%26fulltextType%3DRA%26fileId%3DS1946427400119281.html',
    ... u'filename': '{}/test_data_json_stream.json'.format(test_data_path),
    ... u'vol': u'879',
    ... u'authors': [u'Humberto A. Monreala', u' Alberto M. Villafa\xf1e',
    ...              u' Jos\xe9 G. Chac\xf3n', u' Perla E. Garc\xeda',
    ...              u'Carlos A. Mart\xednez'],
    ... u'year': u'1917'}
    True

    """
    import json

    with open(filename, 'r') as f:
        for n, line in enumerate(f):
            try:
                output = json.loads(line)
                output["filename"] = filename
                yield output
            except ValueError as e:
                logging.debug("Unable to process line: {} (error was: {})".format(str(line), e))
                raise


def __is_iterable(obj):
    try:
        iter(obj)
    except TypeError as te:
        return False
    return True


def _test_json_input(filename):
    """This is here to test the first item of the JSON stream, so that we raise an exception with
    the _iter_json_stream reader and fall back to _iter_large_json."""
    return next(_iter_document_json_stream(filename))


def _iter_large_json(filename, json_prefix='item'):
    # TODO: add the script to automatically find the json_prefix based on a key
    # Also should still have the option to manually specify a prefix for complex
    # json structures.
    """Iterate over all items and sub-items in a json object that match the specified prefix


    Parameters
    ----------
    filename : str
        The filename of the large json file

    json_prefix : str
        The string representation of the hierarchical prefix where the items of 
        interest may be located within the larger json object.

        Try the following script if you need help determining the desired prefix:
        $   import ijson
        $       with open('test_data_large_json_2.json', 'r') as f:
        $           parser = ijson.parse(f)
        $           for prefix, event, value in parser:
        $               print("prefix = '%r' || event = '%r' || value = '%r'" %
        $                     (prefix, event, value))


    Examples
    --------
    >>> documents = _iter_large_json(
    ...             '{}/test_data_large_json.json'.format(test_data_path),
    ...             json_prefix='item._source.isAuthorOf')
    >>> next(documents) == {
    ... u'a': u'ScholarlyArticle',
    ... u'name': u'Path planning and formation control via potential function for UAV Quadrotor',
    ... u'author': [
    ...     u'http://dig.isi.edu/autonomy/data/author/a.a.a.rizqi',
    ...     u'http://dig.isi.edu/autonomy/data/author/t.b.adji',
    ...     u'http://dig.isi.edu/autonomy/data/author/a.i.cahyadi'],
    ... u'text': u"Potential-function-based control strategy for path planning and formation " +
    ...     u"control of Quadrotors is proposed in this work. The potential function is " +
    ...     u"used to attract the Quadrotor to the goal location as well as avoiding the " +
    ...     u"obstacle. The algorithm to solve the so called local minima problem by utilizing " +
    ...     u"the wall-following behavior is also explained. The resulted path planning via " +
    ...     u"potential function strategy is then used to design formation control algorithm. " +
    ...     u"Using the hybrid virtual leader and behavioral approach schema, the formation " +
    ...     u"control strategy by means of potential function is proposed. The overall strategy " +
    ...     u"has been successfully applied to the Quadrotor's model of Parrot AR Drone 2.0 in " +
    ...     u"Gazebo simulator programmed using Robot Operating System.\\nAuthor(s) Rizqi, A.A.A. " +
    ...     u"Dept. of Electr. Eng. & Inf. Technol., Univ. Gadjah Mada, Yogyakarta, Indonesia " +
    ...     u"Cahyadi, A.I. ; Adji, T.B.\\nReferenced Items are not available for this document.\\n" +
    ...     u"No versions found for this document.\\nStandards Dictionary Terms are available to " +
    ...     u"subscribers only.",
    ... u'uri': u'http://dig.isi.edu/autonomy/data/article/6871517',
    ... u'datePublished': u'2014',
    ... 'filename': '{}/test_data_large_json.json'.format(test_data_path)}
    True
    """

    from ijson import items
    with open(filename, 'r') as f:
        for item in items(f, json_prefix):
            if hasattr(item, 'keys'): # check if item is a dictionary
                item['filename'] = filename
                yield item
            # check if item is both iterable and not a string
            elif __is_iterable(item) and not isinstance(item, str):
                for sub_item in item:
                    # check if sub_item is a dictionary
                    if hasattr(sub_item, 'keys'):
                        sub_item['filename'] = filename
                        yield sub_item
            else:
                raise ValueError("'item' in json source is not a dict, and is either a string or not iterable: %r" % item)


def _iter_documents_folder(folder, content_field='text'):
    """Iterate over the files in a folder to retrieve the content to process and tokenize.

    Parameters
    ----------
    folder : str
        The folder containing the files you want to analyze.

    content_field : str
        The usage of 'content_field' in this source is different from most other sources.  The 
        assumption in this source is that each file contains raw text, NOT dictionaries of 
        categorized data.  The content_field argument here specifies what key to store the raw
        text under in the returned dictionary for each document.

    Examples
    --------
    >>> documents = _iter_documents_folder(
    ...     '{}/test_data_folder_files'.format(test_data_path))
    >>> next(documents)['text'] == (
    ...     u"'Interstellar' was incredible. The visuals, the score, " +
    ...     u"the acting, were all amazing. The plot is definitely one " +
    ...     u"of the most original I've seen in a while.")
    True
    """

    import gzip

    if not os.path.exists(folder):
        raise IOError("Folder not found!")

    for directory, subdirectories, files in os.walk(folder):
        for n, file in enumerate(sorted(files)):
            _open = gzip.open if file.endswith('.gz') else open
            try:
                fullpath = os.path.join(directory, file)
                with _open(fullpath, 'rb') as f:
                    yield {content_field: f.read().decode('utf-8'),
                           'filename': fullpath}
            except (ValueError, UnicodeDecodeError) as err:
                logging.warning("Unable to process file: {}, error: {}".format(fullpath, err))


def _iter_solr_query(solr_instance, content_field, query="*:*", content_in_list=True, **kwargs):
    # TODO: should I be checking for presence of content_field and year_field?
    # If not then we don't need them as params
    # TODO: should I care at this level whether the content_in_list or only at
    # the read_input level?  Is it being handled at that level currently?
    """Iterate over all documents in the specified solr intance that match the specified query

    Parameters
    ----------
    solr_instance : str
        Address of the solr instance
    content_field : str
        The name fo the field that contains the main text body of the document.
    query : str
        The solr query string
    content_in_list: bool
        Whether the source fields are stored in single-element lists.  Used for unpacking.
    """

    import pysolr

    s = pysolr.Solr(solr_instance)
    results_per_batch = 100
    response = s.select(query, rows=results_per_batch)
    while response.results:
        for item in response.results:
            if content_in_list:
                if content_field in item.keys() and type(item[content_field]) == list:
                    item[content_field] = item[content_field][0]
            yield item
        response = response.next_batch()


def _iter_elastic_query(hosts, **kwargs):
    """Iterate over all documents in the specified elasticsearch intance and index that match the specified query.

    kwargs are passed to Elasticsearch class instantiation, and can be used to pass any additional options
    described at https://elasticsearch-py.readthedocs.org/en/master/

    Parameters
    ----------
    hosts : str or list
        Address of the elasticsearch instance any index.  May include port, username and password.
        See https://elasticsearch-py.readthedocs.org/en/master/api.html#elasticsearch for all options.

    content_field : str
        The name fo the field that contains the main text body of the document.

    **kwargs: additional keyword arguments to be passed to Elasticsearch client instance and to scan query.
              See
              https://elasticsearch-py.readthedocs.org/en/master/api.html#elasticsearch for all client options.
              https://elasticsearch-py.readthedocs.org/en/master/helpers.html#elasticsearch.helpers.scan for all scan options.
    """

    from elasticsearch import Elasticsearch, helpers
    es = Elasticsearch(hosts, **kwargs)
    results = helpers.scan(es, **kwargs)
    for result in results:
        yield result['_source']


def read_input(source, content_field, source_type="auto",
               output_type=DictionaryCorpus.class_key(), output_args=None,
               synchronous_wait=0, **kwargs):
    """
    Read data from given source into Topik's internal data structures.

    Parameters
    ----------
    source : str
        input data.  Can be file path, directory, or server address.
    content_field : str
        Which field contains your data to be analyzed.  Hash of this is used as id.
    source_type : str
        "auto" tries to figure out data type of source.  Can be manually specified instead.
        options for manual specification are ['solr', 'elastic', 'json_stream', 'large_json', 'folder']
    output_type : str
        Internal format for handling user data.  Current options are in the registered_outputs dictionary.
        Default is DictionaryCorpus class.  Specify alternatives using string key from dictionary.
    output_args : dict
        Configuration to pass through to output
    synchronous_wait : positive, real number
        Time in seconds to wait for data to finish uploading to output (this happens
        asynchronously.)  Only relevant for some output types ("elastic", not "dictionary")
    kwargs : any other arguments to pass to input parsers

    Returns
    -------
    iterable output object


    Examples
    --------
    >>> raw_data = read_input(
    ...         '{}/test_data_json_stream.json'.format(test_data_path),
    ...          content_field="abstract")
    >>> id, text = next(iter(raw_data))
    >>> text == (
    ... u'Transition metal oxides are being considered as the next generation '+
    ... u'materials in field such as electronics and advanced catalysts; '+
    ... u'between them is Tantalum (V) Oxide; however, there are few reports '+
    ... u'for the synthesis of this material at the nanometer size which could '+
    ... u'have unusual properties. Hence, in this work we present the '+
    ... u'synthesis of Ta2O5 nanorods by sol gel method using DNA as structure '+
    ... u'directing agent, the size of the nanorods was of the order of 40 to '+
    ... u'100 nm in diameter and several microns in length; this easy method '+
    ... u'can be useful in the preparation of nanomaterials for electronics, '+
    ... u'biomedical applications as well as catalysts.')
    True
    """

    import time
    json_extensions = [".js", ".json"]
    if output_args is None:
        output_args = {}

    # solr defaults to port 8983
    if (source_type == "auto" and "8983" in source) or source_type == "solr":
        data_iterator = _iter_solr_query(source, **kwargs)
    # web addresses default to elasticsearch
    elif (source_type == "auto" and "9200" in source) or source_type == "elastic":
        data_iterator = _iter_elastic_query(source, **kwargs)
    # files must end in .json.  Try json parser first, try large_json parser next.  Fail otherwise.
    elif (source_type == "auto" and os.path.splitext(source)[1] in json_extensions) or source_type == "json_stream":
        try:
            _test_json_input(source)
            data_iterator = _iter_document_json_stream(source, **kwargs)
        except ValueError as e:
            data_iterator = _iter_large_json(source, **kwargs)
    elif source_type == "large_json":
        data_iterator = _iter_large_json(source, **kwargs)
    # folder paths are simple strings that don't end in an extension (.+3-4 characters), or end in a /
    elif (source_type == "auto" and os.path.splitext(source)[1] == "") or source_type == "folder":
        data_iterator = _iter_documents_folder(source, content_field=content_field)
    else:
        raise ValueError("Unrecognized source type: {}.  Please either manually specify the type, or convert your input"
                         " to a supported type.".format(source))
    if "content_field" not in output_args:
        output_args["content_field"] = content_field
    output = registered_outputs[output_type](iterable=data_iterator, **output_args)

    if synchronous_wait > 0:
        start = time.time()
        items_stored = len(output)
        time.sleep(1)
        # TODO: might get into trouble here if upload is actually empty!
        while len(output) <= 0 or (items_stored != len(output) and time.time() - start < synchronous_wait):
            logging.debug("Number of documents added to the index: {}".format(len(output)))
            time.sleep(1)
            items_stored = len(output)
        if time.time() - start > synchronous_wait:
            logging.warning("Number of documents had not stabilized by end of synchronous_wait, continuing anyway.")
    return output



import click

from topik.run import run_model


@click.command(help='Run topic modeling')
@click.option("-d", "--data", required=True, help="Path to input data for topic modeling")
@click.option("-c", "--field", help="the content field to extract text from, or for folders,"
              " the field to store text as", required=True)
@click.option("-f", "--format", default="auto", help="Data format provided: "
                 "json_stream, folder_files, large_json, solr, elastic")
@click.option("-m", "--model", help="Statistical topic model: lda, plsa", default="LDA")
@click.option("-o", "--output", help="Topic modeling output path", default="./topic_model")
@click.option("-t", "--tokenizer", help="Tokenize method to use: "
                "simple, collocations, entities, mix", default='simple')
@click.option("-n", "--ntopics", help="Number of topics to find", default=10)
@click.option("--termite", help="Whether to output a termite plot as a result", default=True)
@click.option("--ldavis", help="Whether to output an LDAvis-type plot as a result", default=False)
def run(data, format, output, tokenizer, ntopics, field, model, termite, ldavis):
    run_model(data_source=data, source_type=format, dir_path=output, tokenizer=tokenizer, n_topics=ntopics,
              content_field=field, model=model, r_ldavis=ldavis, termite_plot=termite)



# -*- coding: utf-8 -*-

import logging
import math
import operator
import random

import numpy as np
import pandas as pd

from .model_base import TopicModelBase, register_model
from topik.intermediaries.raw_data import load_persisted_corpus


logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                    level=logging.INFO)


# def _rand_mat(sizex, sizey):
#     ret = []
#     for i in xrange(sizex):
#         ret.append([])
#         for _ in xrange(sizey):
#             ret[-1].append(random.random())
#         norm = sum(ret[-1])
#         for j in xrange(sizey):
#             ret[-1][j] /= norm
#     return ret

def _rand_mat(cols, rows):
    out = np.random.random((rows, cols))
    for row in out:
        row /= row.sum()
    return out


@register_model
class PLSA(TopicModelBase):
    def __init__(self, corpus=None, ntopics=2, load_filename=None, binary_filename=None):
        # corpus comes in as a list of lists of tuples.  Each inner list represents a document, while each
        #     tuple contains (id, count) of words in that document.
        self.topics = ntopics
        self.topic_array = np.arange(ntopics, dtype=np.int32)
        if corpus:
            # iterable, each entry is tuple of (word_id, count)
            self._corpus = corpus
            # total number of identified words for each given document (document length normalization factor?)
            self.each = map(sum, map(lambda x: x[1], corpus))
            # Maximum identified word (number of identified words in corpus)
            # TODO: seems like this could be tracked better during the tokenization step and fed in.
            self.words = len(corpus._dict.token2id)
            self.likelihood = 0
            # topic-word matrix
            self.zw = _rand_mat(self.words, self.topics)
            # document-topic matrix
            self.dz = _rand_mat(self.topics, len(corpus))
            self.dw_z = [{}, ] * len(corpus)
            self.p_dw = [{}, ] * len(corpus)
            self.beta = 0.8
        elif load_filename and binary_filename:
            from topik.intermediaries.digested_document_collection import DigestedDocumentCollection
            self._corpus = DigestedDocumentCollection(load_persisted_corpus(load_filename))
            # total number of identified words for each given document (document length normalization factor?)
            self.each = map(sum, map(lambda x: x[1], self._corpus))
            # Maximum identified word (number of identified words in corpus)
            # TODO: seems like this could be tracked better during the tokenization step and fed in.
            self.words = max(reduce(operator.add, map(lambda x: x[0], self._corpus)))+1
            arrays = np.load(binary_filename)
            self.zw = arrays['zw']
            self.dz = arrays['dz']
            self.dw_z = arrays['dw_z']
            self.p_dw = arrays['p_dw']
            self.beta, self.likelihood = arrays["beta_likelihood"]
        else:
            pass  # is just being used for inference

    def save(self, filename):
        np.savez_compressed(self.get_model_name_with_parameters(),
                            zw=self.zw,
                            dz=self.dz,
                            dw_z=self.dw_z,
                            p_dw=self.p_dw,
                            beta_likelihood=np.array([self.beta, self.likelihood]))
        saved_data = {"load_filename": filename, "binary_filename": self.get_model_name_with_parameters()+".npz"}
        super(PLSA, self).save(filename, saved_data=saved_data)

    def get_model_name_with_parameters(self):
        return "PLSA_{}_topics{}".format(self.topics, self._corpus.filter_string)

    def _cal_p_dw(self):
        for d, doc in enumerate(self._corpus):
            for word_id, word_ct in doc:
                tmp = 0
                for _ in range(word_ct):
                    for z in self.topic_array:
                        tmp += (self.zw[z][word_id]*self.dz[d][z])**self.beta
                self.p_dw[-1][word_id] = tmp

    def _e_step(self):
        self._cal_p_dw()
        for d, doc in enumerate(self._corpus):
            for word_id, word_ct in doc:
                self.dw_z[-1][word_id] = []
                for z in self.topic_array:
                    self.dw_z[-1][word_id].append(((self.zw[z][word_id]*self.dz[d][z])**self.beta)/self.p_dw[d][word_id])

    def _m_step(self):
        for z in self.topic_array:
            self.zw[z] = [0]*self.words
            for d, doc in enumerate(self._corpus):
                for word_id, word_ct in doc:
                    self.zw[z][word_id] += word_ct*self.dw_z[d][word_id][z]
            norm = sum(self.zw[z])
            for w in xrange(self.words):
                self.zw[z][w] /= norm
        for d, doc in enumerate(self._corpus):
            self.dz[d] = 0
            for z in self.topic_array:
                for word_id, word_ct in doc:
                    self.dz[d][z] += word_ct * self.dw_z[d][word_id][z]
            for z in self.topic_array:
                self.dz[d][z] /= self.each[d]

    def _cal_likelihood(self):
        self.likelihood = 0
        for d, doc in enumerate(self._corpus):
            for word_id, word_ct in doc:
                self.likelihood += word_ct*math.log(self.p_dw[d][word_id])

    def train(self, max_iter=100):
        cur = 0
        for i in xrange(max_iter):
            logging.info('%d iter' % i)
            self._e_step()
            self._m_step()
            self._cal_likelihood()
            logging.info('likelihood %f ' % self.likelihood)
            if cur != 0 and abs((self.likelihood-cur)/cur) < 1e-8:
                break
            cur = self.likelihood

    def inference(self, doc, max_iter=100):
        doc = dict(filter(lambda x: x[0] < self.words, doc.items()))
        words = sum(doc.values())
        ret = []
        for i in xrange(self.topics):
            ret.append(random.random())
        norm = sum(ret)
        for i in xrange(self.topics):
            ret[i] /= norm
        tmp = 0
        for _ in xrange(max_iter):
            p_dw = {}
            for w in doc:
                p_dw[w] = 0
                for _ in range(doc[w]):
                    for z in xrange(self.topics):
                        p_dw[w] += (ret[z]*self.zw[z][w])**self.beta
            # e setp
            dw_z = {}
            for w in doc:
                dw_z[w] = []
                for z in xrange(self.topics):
                    dw_z[w].append(((self.zw[z][w]*ret[z])**self.beta)/p_dw[w])
            logging.debug('inference dw_z %r' % (dw_z,))
            # m step

            ret = [0]*self.topics
            for z in xrange(self.topics):
                for w in doc:
                    ret[z] += doc[w]*dw_z[w][z]
            for z in xrange(self.topics):
                ret[z] /= words
            # cal likelihood
            likelihood = 0
            for w in doc:
                likelihood += doc[w]*math.log(p_dw[w])
            if tmp != 0 and abs((likelihood-tmp)/tmp) < 1e-8:
                break
            tmp = likelihood
        return (ret, likelihood)

    def post_prob_sim(self, docd, q):
        sim = 0
        for w in docd:
            tmp = 0
            for z in xrange(self.topics):
                tmp += self.zw[z][w]*q[z]
            sim += docd[w]*math.log(tmp)
        return sim

    def get_top_words(self, topn):
        top_words = []
        # each "topic" is a row of the dz matrix
        for topic in self.dz.T:
            word_ids = np.argpartition(topic, -topn)[-topn:]
            word_ids = reversed(word_ids[np.argsort(topic[word_ids])])
            top_words.append([(topic[word_id], self._corpus.get_id2word_dict()[word_id]) for word_id in word_ids])
        return top_words

    def _get_topic_term_dists(self):
        term_topic_df = pd.DataFrame(self.zw,
                            index=['topic'+str(t)+'dist' for t in range(self.topics)]).T

        term_topic_df.index.name = 'term_id'
        return term_topic_df

    def _get_doc_topic_dists(self):
        doc_topic_df = pd.DataFrame(self.dz,
                            index=[doc[0] for doc in self._corpus._corpus],
                            columns=['topic'+str(t)+'dist' for t in range(self.topics)])

        doc_topic_df.index.name = 'doc_id'
        return doc_topic_df



from __future__ import absolute_import, print_function

import logging
import os
import subprocess
import time
import webbrowser

import numpy as np

from topik.readers import read_input
import topik.models
from topik.viz import plot_lda_vis, Termite


logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                    level=logging.INFO)

BASEDIR = os.path.abspath(os.path.dirname(__file__))


def run_model(data_source, source_type="auto", year_field=None, start_year=None, stop_year=None,
              content_field=None, tokenizer='simple', n_topics=10, dir_path='./topic_model', model='LDA',
              termite_plot=True, output_file=False, ldavis=False, seed=42, **kwargs):

    """Run your data through all topik functionality and save all results to a specified directory.

    Parameters
    ----------
    data_source : str
        Input data (e.g. file or folder or solr/elasticsearch instance).

    source_type : {'json_stream', 'folder_files', 'json_large', 'solr', 'elastic'}.
        The format of your data input. Currently available a json stream or a folder containing text files.
        Default is 'json_stream'
    year_field : str
        The field name (if any) that contains the year associated with each document (for filtering).
    start_year : int
        For beginning of range filter on year_field values
    stop_year : int
        For beginning of range filter on year_field values
    content_field : string
        The primary text field to parse.
    tokenizer : {'simple', 'collocations', 'entities', 'mixed'}
        The type of tokenizer to use. Default is 'simple'.
    n_topics : int
        Number of topics to find in your data
    dir_path : str
        Directory path to store all topic modeling results files. Default is `./topic_model`.
    model : {'LDA', 'PLSA'}.
        Statistical modeling algorithm to use. Default 'LDA'.
    termite_plot : bool
        Generate termite plot of your model if True. Default is True.
    output_file : bool
        Generate a final summary csv file of your results. For each document: text, tokens, lda_probabilities and topic.
    ldavis : bool
        Generate an interactive data visualization of your topics. Default is False.
    seed : int
        Set random number generator to seed, to be able to reproduce results. Default 42.
    **kwargs : additional keyword arguments, passed through to each individual step
    """

    np.random.seed(seed)

    raw_data = read_input(data_source, content_field=content_field,
                          source_type=source_type, **kwargs)
    processed_data = raw_data.tokenize(method=tokenizer, **kwargs)
    model = topik.models.registered_models[model](processed_data, n_topics, **kwargs)
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)

    if termite_plot:
        termite = Termite(model.termite_data(n_topics), "Termite Plot")
        termite.plot(os.path.join(dir_path, 'termite.html'))

    if ldavis:
        plot_lda_vis(model.to_py_lda_vis())





from .readers import read_input
from .models import registered_models, load_model
from .intermediaries.raw_data import registered_outputs, load_persisted_corpus

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import unittest

from topik.models import registered_models, PLSA
from topik.readers import read_input
from topik.tests import test_data_path


class TestPLSA(unittest.TestCase):
    def setUp(self):
        test_raw_data = read_input('{}/test_data_json_stream.json'.format(
            test_data_path), content_field="abstract")
        self.processed_data = test_raw_data.tokenize()

    def test_train(self):
        plsa = PLSA(self.processed_data, ntopics=5)
        plsa.train()
        assert(plsa.likelihood != 0)  # TODO: this is a really poor test, but at least makes sure that things run.

    def test_registration(self):
        assert("PLSA" in registered_models)



import unittest

import nose.tools as nt
import elasticsearch

from topik.readers import read_input
import topik.readers as readers
from topik.intermediaries.raw_data import ElasticSearchCorpus
from topik.tests import test_data_path

INDEX = "test_elastic"

class TestReader(unittest.TestCase):

    def setUp(self):
        self.solution_1 = str("'Interstellar' was incredible. The visuals, the score, the acting, were all amazing."
                              " The plot is definitely one of the most original I've seen in a while.")
        self.solution_2 = str('In this paper, we describe a simple and rapid solution-phase chemical reduction method'
                              ' with no inert gas protection, for preparing stable copper nanoparticle colloid with'
                              ' average particle size of 3.4 nm and narrow size distribution. In our synthesis route,'
                              ' ascorbic acid, natural vitamin C (VC), serves as both a reducing agent and an'
                              ' antioxidant to reduce copper salt precursor and effectively prevent the general'
                              ' oxidation process occurring to the newborn nanoparticles. XRD and UV/vis confirm the'
                              ' formation of pure face-centered cubic (fcc) copper nanoparticles and the excellent'
                              ' antioxidant ability of ascorbic acid.')
        self.solution_3 =   "Potential-function-based control strategy for path planning and formation control of Quadrotors is proposed in this work. The potential function is used to attract the Quadrotor to the goal location as well as avoiding the obstacle. The algorithm to solve the so called local minima problem by utilizing the wall-following behavior is also explained. The resulted path planning via potential function strategy is then used to design formation control algorithm. Using the hybrid virtual leader and behavioral approach schema, the formation control strategy by means of potential function is proposed. The overall strategy has been successfully applied to the Quadrotor's model of Parrot AR Drone 2.0 in Gazebo simulator programmed using Robot Operating System.\nAuthor(s) Rizqi, A.A.A. Dept. of Electr. Eng. & Inf. Technol., Univ. Gadjah Mada, Yogyakarta, Indonesia Cahyadi, A.I. ; Adji, T.B.\nReferenced Items are not available for this document.\nNo versions found for this document.\nStandards Dictionary Terms are available to subscribers only."
        '''(u"Potential-function-based control strategy for path planning and formation control of" +
                            u"Quadrotors is proposed in this work. The potential function is used to attract the " +
                            u"Quadrotor to the goal location as well as avoiding the obstacle. The algorithm to " +
                            u"solve the so called local minima problem by utilizing the wall-following behavior " +
                            u"is also explained. The resulted path planning via potential function strategy is " +
                            u"then used to design formation control algorithm. Using the hybrid virtual leader and " +
                            u"behavioral approach schema, the formation control strategy by means of potential " +
                            u"function is proposed. The overall strategy has been successfully applied to the " +
                            u"Quadrotor's model of Parrot AR Drone 2.0 in Gazebo simulator programmed using Robot " +
                            u"Operating System.\nAuthor(s) Rizqi, A.A.A. Dept. of Electr. Eng. & Inf. Technol., " +
                            u"Univ. Gadjah Mada, Yogyakarta, Indonesia Cahyadi, A.I. ; Adji, T.B.\nReferenced Items " +
                            u"are not available for this document.\nNo versions found for this document.\nStandards " +
                            u"Dictionary Terms are available to subscribers only.")'''
        #self.solution_4 = 'Transition metal oxides are being considered as the next generation materials in field such as electronics and advanced catalysts; between them is Tantalum (V) Oxide; however, there are few reports for the synthesis of this material at the nanometer size which could have unusual properties. Hence, in this work we present the synthesis of Ta2O5 nanorods by sol gel method using DNA as structure directing agent, the size of the nanorods was of the order of 40 to 100 nm in diameter and several microns in length; this easy method can be useful in the preparation of nanomaterials for electronics, biomedical applications as well as catalysts.'
        self.solution_4 = str("Transition metal oxides are being considered as the next generation materials in field "
                              "such as electronics and advanced catalysts; between them is Tantalum (V) Oxide; however"
                              ", there are few reports for the synthesis of this material at the nanometer size which "
                              "could have unusual properties. Hence, in this work we present the synthesis of Ta2O5 "
                              "nanorods by sol gel method using DNA as structure directing agent, the size of the "
                              "nanorods was of the order of 40 to 100 nm in diameter and several microns in length; "
                              "this easy method can be useful in the preparation of nanomaterials for electronics, "
                              "biomedical applications as well as catalysts.")
        '''str(  u"Transition metal oxides are being considered as the next generation materials in field"
                                u"such as electronics and advanced catalysts; between them is Tantalum (V) Oxide; "
                                u"however, there are few reports for the synthesis of this material at the nanometer "
                                u"size which could have unusual properties. Hence, in this work we present the "
                                u"synthesis of Ta2O5 nanorods by sol gel method using DNA as structure directing agent, "
                                u"the size of the nanorods was of the order of 40 to 100 nm in diameter and several "
                                u"microns in length; this easy method can be useful in the preparation of nanomaterials "
                                u"for electronics, biomedical applications as well as catalysts.")'''

    def tearDown(self):
        instance = elasticsearch.Elasticsearch("localhost")
        if instance.indices.exists(INDEX):
            instance.indices.delete(INDEX)

    def test_read_document_json_stream(self):
        iterable_data = read_input('{}/test_data_json_stream.json'.format(
                                   test_data_path), content_field="abstract")
        id, first_text = next(iter(iterable_data))
        self.assertEqual(first_text, self.solution_4)

    def test_read_documents_folder(self):
        loaded_dictionaries = read_input(
            '{}/test_data_folder_files'.format(test_data_path),
            content_field="abstract")
        id, first_text = next(iter(loaded_dictionaries))
        self.assertEqual(first_text, self.solution_1)

    def test_iter_documents_folder_gz(self):
        loaded_dictionaries = read_input(
            '{}/test_data_folder_files_gz'.format(test_data_path),
            content_field="abstract")
        id, first_text = next(iter(loaded_dictionaries))
        self.assertEqual(first_text, self.solution_1)

    def test_iter_large_json(self):
        iterable_data = read_input('{}/test_data_large_json.json'.format(test_data_path),
                                   content_field="text", json_prefix='item._source.isAuthorOf')
        id, first_text = next(iter(iterable_data))
        self.assertEqual(first_text, self.solution_3)

    def test_elastic_import(self):
        output_args = {'source': 'localhost',
                       'index': INDEX}
        # import data from file into known elastic server
        read_input('{}/test_data_json_stream.json'.format(
                   test_data_path), content_field="abstract",
                   output_type=ElasticSearchCorpus.class_key(),
                   output_args=output_args, synchronous_wait=30)
        iterable_data = read_input("localhost", source_type="elastic", content_field="abstract", index=INDEX)
        self.assertEquals(len(iterable_data), 100)


def test_bad_folder():
    nt.assert_raises(IOError, next, readers._iter_documents_folder("Frank"))


if __name__ == '__main__':
    unittest.main()



"""
This file is concerned with providing a simple interface for data stored in
Elasticsearch.  The class(es) defined here are fed into the preprocessing step.
"""

from abc import ABCMeta, abstractmethod, abstractproperty
import logging
import time

from six import with_metaclass

from topik.intermediaries.persistence import Persistor
from topik.tokenizers import tokenizer_methods
from topik.intermediaries.digested_document_collection import DigestedDocumentCollection


registered_outputs = {}

def register_output(cls):
    global registered_outputs
    if cls.class_key() not in registered_outputs:
        registered_outputs[cls.class_key()] = cls
    return cls


def _get_hash_identifier(input_data, id_field):
    return hash(input_data[id_field])


def _get_parameters_string(**kwargs):
    """Used to create identifiers for output"""
    id = ''.join('{}={}_'.format(key, val) for key, val in sorted(kwargs.items()))
    return id[:-1]


class CorpusInterface(with_metaclass(ABCMeta)):
    def __init__(self):
        super(CorpusInterface, self).__init__()
        self.persistor = Persistor()

    @classmethod
    @abstractmethod
    def class_key(cls):
        """Implement this method to return the string ID with which to store your class."""
        raise NotImplementedError

    @abstractmethod
    def __iter__(self):
        """This is expected to iterate over your data, returning tuples of (doc_id, <selected field>)"""
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def get_generator_without_id(self, field=None):
        """Returns a generator that yields field content without doc_id associate"""
        raise NotImplementedError

    @abstractmethod
    def append_to_record(self, record_id, field_name, field_value):
        """Used to store preprocessed output alongside input data.

        Field name is destination.  Value is processed value."""
        raise NotImplementedError

    @abstractmethod
    def get_date_filtered_data(self, start, end, field):
        raise NotImplementedError

    @abstractproperty
    def filter_string(self):
        raise NotImplementedError

    def save(self, filename, saved_data=None):
        """Persist this object to disk somehow.

        You can save your data in any number of files in any format, but at a minimum, you need one json file that
        describes enough to bootstrap the loading prcess.  Namely, you must have a key called 'class' so that upon
        loading the output, the correct class can be instantiated and used to load any other data.  You don't have
        to implement anything for saved_data, but it is stored as a key next to 'class'.

        """
        self.persistor.store_corpus({"class": self.__class__.class_key(), "saved_data": saved_data})
        self.persistor.persist_data(filename)

    def synchronize(self, max_wait, field):
        """By default, operations are synchronous and no additional wait is
        necessary.  Data sources that are asynchronous (ElasticSearch) may
        use this function to wait for "eventual consistency" """
        pass

    def tokenize(self, method="simple", synchronous_wait=30, **kwargs):
        """Convert data to lowercase; tokenize; create bag of words collection.

        Output from this function is used as input to modeling steps.

        raw_data: iterable corpus object containing the text to be processed.
            Each iteration call should return a new document's content.
        tokenizer_method: string id of tokenizer to use.  For keys, see
            topik.tokenizers.tokenizer_methods (which is a dictionary of classes)
        kwargs: arbitrary dicionary of extra parameters.  These are passed both
            to the tokenizer and to the vectorizer steps.
        """
        parameters_string = _get_parameters_string(method=method, **kwargs)
        token_path = "tokens_"+parameters_string
        for record_id, raw_record in self:
            tokenized_record = tokenizer_methods[method](raw_record,
                                         **kwargs)
            # TODO: would be nice to aggregate batches and append in bulk
            self.append_to_record(record_id, token_path, tokenized_record)
        self.synchronize(max_wait=synchronous_wait, field=token_path)
        return DigestedDocumentCollection(self.get_field(field=token_path))


@register_output
class ElasticSearchCorpus(CorpusInterface):
    def __init__(self, source, index, content_field, doc_type=None, query=None, iterable=None,
                 filter_expression="", **kwargs):
        from elasticsearch import Elasticsearch
        super(ElasticSearchCorpus, self).__init__()
        self.hosts = source
        self.instance = Elasticsearch(hosts=source, **kwargs)
        self.index = index
        self.content_field = content_field
        self.doc_type = doc_type
        self.query = query
        if iterable:
            self.import_from_iterable(iterable, content_field)
        self.filter_expression = filter_expression

    @classmethod
    def class_key(cls):
        return "elastic"

    @property
    def filter_string(self):
        return self.filter_expression

    def __iter__(self):
        from elasticsearch import helpers
        results = helpers.scan(self.instance, index=self.index,
                               query=self.query, doc_type=self.doc_type)
        for result in results:
            yield result["_id"], result['_source'][self.content_field]

    def __len__(self):
        return self.instance.count(index=self.index, doc_type=self.doc_type)["count"]

    def get_generator_without_id(self, field=None):
        if not field:
            field = self.content_field
        for (_, result) in ElasticSearchCorpus(self.hosts, self.index, field, self.doc_type, self.query):
            yield result

    def append_to_record(self, record_id, field_name, field_value):
        self.instance.update(index=self.index, id=record_id, doc_type="continuum",
                             body={"doc": {field_name: field_value}})

    def get_field(self, field=None):
        """Get a different field to iterate over, keeping all other
           connection details."""
        if not field:
            field = self.content_field
        return ElasticSearchCorpus(self.hosts, self.index, field, self.doc_type, self.query)

    def import_from_iterable(self, iterable, id_field="text", batch_size=500):
        """Load data into Elasticsearch from iterable.

        iterable: generally a list of dicts, but possibly a list of strings
            This is your data.  Your dictionary structure defines the schema
            of the elasticsearch index.
        id_field: string identifier of field to hash for content ID.  For
            list of dicts, a valid key value in the dictionary is required. For
            list of strings, a dictionary with one key, "text" is created and
            used.
        """
        from elasticsearch import helpers
        batch = []
        for item in iterable:
            if isinstance(item, basestring):
                item = {id_field: item}
            id = _get_hash_identifier(item, id_field)
            action = {'_op_type': 'update',
                      '_index': self.index,
                      '_type': 'continuum',
                      '_id': id,
                      'doc': item,
                      'doc_as_upsert': "true",
                      }
            batch.append(action)
            if len(batch) >= batch_size:
                helpers.bulk(client=self.instance, actions=batch, index=self.index)
                batch = []
        if batch:
            helpers.bulk(client=self.instance, actions=batch, index=self.index)
        self.instance.indices.refresh(self.index)

    def convert_date_field_and_reindex(self, field):
        index = self.index
        if self.instance.indices.get_field_mapping(field=field,
                                           index=index,
                                           doc_type="continuum") != 'date':
            index = self.index+"_{}_alias_date".format(field)
            if not self.instance.indices.exists(index) or self.instance.indices.get_field_mapping(field=field,
                                           index=index,
                                           doc_type="continuum") != 'date':
                mapping = self.instance.indices.get_mapping(index=self.index,
                                                            doc_type="continuum")
                mapping[self.index]["mappings"]["continuum"]["properties"][field] = {"type": "date"}
                self.instance.indices.put_alias(index=self.index,
                                                name=index,
                                                body=mapping)
                self.instance.indices.refresh(index)
                while self.instance.count(index=self.index) != self.instance.count(index=index):
                    logging.info("Waiting for date indexed data to be indexed...")
                    time.sleep(1)
        return index

    # TODO: validate input data to ensure that it has valid year data
    def get_date_filtered_data(self, start, end, field="date"):
        converted_index = self.convert_date_field_and_reindex(field=field)
        return ElasticSearchCorpus(self.hosts, converted_index, self.content_field, self.doc_type,
                                   query={"query": {"filtered": {"filter": {"range": {field: {"gte": start,
                                                                                              "lte": end}}}}}},
                                   filter_expression=self.filter_expression + "_date_{}_{}".format(start, end))

    def save(self, filename, saved_data=None):
        if saved_data is None:
            saved_data = {"source": self.hosts, "index": self.index, "content_field": self.content_field,
                          "doc_type": self.doc_type, "query": self.query}
        return super(ElasticSearchCorpus, self).save(filename, saved_data)

    def synchronize(self, max_wait, field):
        # TODO: change this to a more general condition for wider use, including read_input
        # could just pass in a string condition and then 'while not eval(condition)'
        count_not_yet_updated = -1
        while count_not_yet_updated != 0:
            count_not_yet_updated = self.instance.count(index=self.index,
                                             doc_type=self.doc_type,
                                             body={"query": {
                                                        "constant_score" : {
                                                            "filter" : {
                                                                "missing" : {
                                                                    "field" : field}}}}})['count']
            logging.debug("Count not yet updated: {}".format(count_not_yet_updated))
            time.sleep(0.01)
        pass

@register_output
class DictionaryCorpus(CorpusInterface):
    def __init__(self, content_field, iterable=None, generate_id=True, reference_field=None, content_filter=None):
        super(DictionaryCorpus, self).__init__()
        self.content_field = content_field
        self._documents = []
        self.idx = 0
        active_field = None
        if reference_field:
            self.reference_field = reference_field
            active_field = content_field
            content_field = reference_field
        else:
            self.reference_field = content_field
        if iterable:
            self.import_from_iterable(iterable, content_field, generate_id)
        if active_field:
            self.content_field = active_field
        self.content_filter = content_filter

    @classmethod
    def class_key(cls):
        return "dictionary"

    def __iter__(self):
        for doc in self._documents:
            if self.content_filter:
                if eval(self.content_filter["expression"].format(doc["_source"][self.content_filter["field"]])):
                    yield doc["_id"], doc["_source"][self.content_field]
            else:
                yield doc["_id"], doc["_source"][self.content_field]

    def __len__(self):
        return len(self._documents)

    @property
    def filter_string(self):
        return self.content_filter["expression"].format(self.content_filter["field"]) if self.content_filter else ""

    def append_to_record(self, record_id, field_name, field_value):
        for doc in self._documents:
            if doc["_id"] == record_id:
                doc["_source"][field_name] = field_value
                return
        raise ValueError("No record with id '{}' was found.".format(record_id))

    def get_field(self, field=None):
        """Get a different field to iterate over, keeping all other details."""
        if not field:
            field = self.content_field
        return DictionaryCorpus(content_field=field, iterable=self._documents,
                                generate_id=False, reference_field=self.content_field)

    def get_generator_without_id(self, field=None):
        if not field:
            field = self.content_field
        for doc in self._documents:
            yield doc["_source"][field]

    def import_from_iterable(self, iterable, content_field, generate_id=True):
        """
        iterable: generally a list of dicts, but possibly a list of strings
            This is your data.  Your dictionary structure defines the schema
            of the elasticsearch index.
        """
        if generate_id:
            self._documents = [{"_id": hash(doc[content_field]),
                                "_source": doc} for doc in iterable]
            self.reference_field = content_field
        else:
            self._documents = [item for item in iterable]

    # TODO: generalize for datetimes
    # TODO: validate input data to ensure that it has valid year data
    def get_date_filtered_data(self, start, end, field="year"):
        return DictionaryCorpus(content_field=field, iterable=self._documents,
                                generate_id=False, reference_field=self.content_field,
                                content_filter={"field": field, "expression": "{}<=int({})<={}".format(start, "{}", end)})

    def save(self, filename, saved_data=None):
        if saved_data is None:
            saved_data = {"reference_field": self.reference_field, "content_field": self.content_field,
                          "iterable": [doc["_source"] for doc in self._documents]}
        return super(DictionaryCorpus, self).save(filename, saved_data)

def load_persisted_corpus(filename):
    corpus_dict = Persistor(filename).get_corpus_dict()
    return registered_outputs[corpus_dict['class']](**corpus_dict["saved_data"])



import gensim

class DigestedDocumentCollection(gensim.interfaces.CorpusABC):
    """A bag-of-words representation of a corpus (collection of documents).

    This serves as direct input to modeling functions.  It is output from
    preprocessing functions.

    Parameters
    ----------
    corpus: A collection of tokenized documents
        Each document is a list of tokens, tokenized and normalized strings
        (either utf8 or unicode) (e.g. output of topik.SimpleTokenizer)

    Readers iterate over tuples (id, content), but discard id in return (for compatibility with Gensim.)

    """
    def __init__(self, tokenized_corpus):
        from gensim.corpora.dictionary import Dictionary
        self._corpus = tokenized_corpus
        self._dict = Dictionary(tokenized_corpus.get_generator_without_id())
        super(DigestedDocumentCollection, self).__init__()

    def __iter__(self):
        """Discards id field - for compatibility with Gensim."""
        for _id, doc_tokens in self._corpus:
            yield self._dict.doc2bow(doc_tokens)

    def __len__(self):
        return len(self._corpus)

    def get_id2word_dict(self):
        return self._dict

    def save(self, filename):
        self._corpus.save(filename)

    @property
    def persistor(self):
        return self._corpus.persistor

    @property
    def filter_string(self):
        return self._corpus.filter_string



from __future__ import absolute_import

import logging

from topik.tests import test_data_path

class Termite(object):
    """A Bokeh Termite Visualization for LDA results analysis.

    Parameters
    ----------
    input_file : str or pandas DataFrame
        A pandas dataframe from a topik model get_termite_data() containing columns "word", "topic" and "weight".
        May also be a string, in which case the string is a filename of a csv file with the above columns.
    title : str
        The title for your termite plot

    Examples
    --------
    >>> termite = Termite("{}/termite.csv".format(test_data_path),
    ...                   "My lda results")
    >>> termite.plot('my_termite.html')

    """
    def __init__(self, input_file, title):
        self.input_file = input_file
        self.title = title

    def plot(self, output_file="termite.html"):
        import blaze as blz
        from odo import into
        import pandas as pd
        import bokeh.plotting as plt
        from bokeh.models.sources import ColumnDataSource

        t = blz.Data(self.input_file)

        MAX = blz.compute(t.weight.max())
        MIN = blz.compute(t.weight.min())

        # Create a size variable to define the size of the the circle for the plot.
        t = blz.transform(t, size=blz.sqrt((t.weight - MIN)/(MAX - MIN))*50)

        WORDS = t['word'].distinct()
        WORDS = into(list, WORDS)
        topics = t['topic'].distinct()
        topics = into(list, topics)
        # Convert topics to strings
        TOPICS = [str(i) for i in topics]

        source = into(pd.DataFrame, t)

        plt.output_file(output_file)

        data_source = ColumnDataSource(source)

        p = plt.figure(x_range=TOPICS, y_range=WORDS,
                       plot_width=1000, plot_height=1700,
                       title=self.title)

        p.circle(x="topic", y="word", size="size", fill_alpha=0.6, source=data_source)
        plt.show(p)

def plot_lda_vis(model_data):
    """Designed to work with to_py_lda_vis() in the model classes."""
    from pyLDAvis import prepare, show
    model_vis_data = prepare(**model_data)
    show(model_vis_data)



import unittest
import os

from topik.viz import Termite

module_path = os.path.dirname(__file__)


class TestTokenizers(unittest.TestCase):

    def test_termite(self):
        termite = Termite(os.path.join(module_path, 'data/termite.csv' ), "My lda results")
        termite.plot(os.path.join(module_path, 'termite_plot.html'))

        self.assertTrue(os.path.isfile(os.path.join(module_path, 'termite_plot.html')))



import os
import unittest

from nose.tools import assert_raises

from topik.readers import read_input
from topik.models import registered_models, load_model
from topik.intermediaries.persistence import Persistor

# sample data files are located in the same folder
module_path = os.path.dirname(__file__)

NTOPICS = 3
MODEL_SAVE_FILENAME = os.path.join(module_path, 'test.model')


class _ModelBase(object):
    def setUp(self):
        raw_data = read_input(
                source=os.path.join(module_path, 'data/test_data_json_stream.json'),
                content_field="abstract")
        self.digested_data = raw_data.tokenize()
        self.model = registered_models[self.model_name](self.digested_data, ntopics=NTOPICS)

    def tearDown(self):
        import glob
        for f in glob.glob(MODEL_SAVE_FILENAME+"*"):
            os.remove(f)

    def test_save_data(self):
        self.model.save(MODEL_SAVE_FILENAME)
        self.assertTrue(os.path.isfile(MODEL_SAVE_FILENAME))
        self.assertTrue("class" in Persistor(MODEL_SAVE_FILENAME).get_corpus_dict())

    def test_load_data(self):
        """NOTE: This test depends on save data succeeding!  May want to have static known-good data instead."""
        self.model.save(MODEL_SAVE_FILENAME)
        model = load_model(MODEL_SAVE_FILENAME, self.model.get_model_name_with_parameters())
        self.assertGreater(model.get_top_words(5), 0)

    def test_top_words(self):
        top_words = self.model.get_top_words(15)
        self.assertEqual(len(top_words), 3)
        self.assertEqual(len(top_words[1]), 15)
        self.assertLess(top_words[0][1][0], top_words[0][0][0])
        self.assertLess(top_words[0][-1][0], top_words[0][-2][0])

    def test_termite_pandas_output(self):
        topn_words = 15
        data = self.model.termite_data(topn_words=topn_words)
        self.assertFalse(data.empty)
        self.assertEqual(len(data.columns), 3)
        self.assertEqual(len(data), topn_words * 3)


class TestLDA(_ModelBase, unittest.TestCase):
    model_name = "LDA"


class TestPLSA(_ModelBase, unittest.TestCase):
    model_name = "PLSA"


def test_invalid_source_load():
    raw_data = read_input(source=os.path.join(module_path, 'data/test_data_json_stream.json'),
                          content_field="abstract")
    raw_data.save("test_file")
    assert_raises(NameError, load_model, "test_file", "Steve")  # attempt to load a model that we know does not exist


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function

import itertools
import logging
import re

# imports used only for doctests
from topik.tests import test_data_path


logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                    level=logging.INFO)


def tokenize_simple(text, min_length=1, stopwords=None):
    """A text tokenizer that simply lowercases, matches alphabetic
    characters and removes stopwords.

    Parameters
    ----------
    text : str
        A single document's text to be tokenized
    entities : iterable of str
        Collection of noun phrases, obtained from collect_entities function
    min_length : int
        Minimum length of any single word
    stopwords: None or iterable of str
        Collection of words to ignore as tokens

    Examples
    --------
    >>> from topik.readers import read_input
    >>> id_documents = read_input(
    ...                 '{}/test_data_json_stream.json'.format(test_data_path),
    ...                 content_field="abstract")
    >>> id, doc_text = next(iter(id_documents))
    >>> doc_text
    u'Transition metal oxides are being considered as the next generation \
materials in field such as electronics and advanced catalysts; between\
 them is Tantalum (V) Oxide; however, there are few reports for the \
synthesis of this material at the nanometer size which could have \
unusual properties. Hence, in this work we present the synthesis of \
Ta2O5 nanorods by sol gel method using DNA as structure directing \
agent, the size of the nanorods was of the order of 40 to 100 nm in \
diameter and several microns in length; this easy method can be useful\
 in the preparation of nanomaterials for electronics, biomedical \
applications as well as catalysts.'
    >>> tokens = tokenize_simple(doc_text)
    >>> tokens
    [u'transition', u'metal', u'oxides', u'considered', \
u'generation', u'materials', u'field', u'electronics', \
u'advanced', u'catalysts', u'tantalum', u'v', u'oxide', \
u'reports', u'synthesis', u'material', u'nanometer', u'size', \
u'unusual', u'properties', u'work', u'present', u'synthesis', \
u'ta', u'o', u'nanorods', u'sol', u'gel', u'method', u'dna', \
u'structure', u'directing', u'agent', u'size', u'nanorods', \
u'order', u'nm', u'diameter', u'microns', u'length', u'easy', \
u'method', u'useful', u'preparation', u'nanomaterials', u'electronics', \
u'biomedical', u'applications', u'catalysts']
    """

    import gensim
    if not stopwords:
        from gensim.parsing.preprocessing import STOPWORDS as stopwords
    return [word for word in gensim.utils.tokenize(text, lower=True)
            if word not in stopwords and len(word) >= min_length]


def collect_bigrams_and_trigrams(collection, top_n=10000, min_length=1, min_bigram_freq=50,
                                 min_trigram_freq=20, stopwords=None):
    """collects bigrams and trigrams from collection of documents.  Input to collocation tokenizer.

    bigrams are pairs of words that recur in the collection; trigrams are triplets.

    Parameters
    ----------
    collection : iterable of str
        body of documents to examine
    top_n : int
        limit results to this many entries
    min_length : int
        Minimum length of any single word
    min_bigram_freq : int
        threshold of when to consider a pair of words as a recognized bigram
    min_trigram_freq : int
        threshold of when to consider a triplet of words as a recognized trigram
    stopwords : None or iterable of str
        Collection of words to ignore as tokens

    Examples
    --------
    >>> from topik.readers import read_input
    >>> raw_data = read_input(
    ...                 '{}/test_data_json_stream.json'.format(test_data_path),
    ...                 content_field="abstract")
    >>> bigrams, trigrams = collect_bigrams_and_trigrams(raw_data, min_bigram_freq=5, min_trigram_freq=3)
    >>> bigrams.pattern
    u'(free standing|ac electrodeposition|centered cubic|spatial resolution|vapor deposition\
|wear resistance|plastic deformation|electrical conductivity|field magnets|v o|\
transmission electron|x ray|et al|ray diffraction|electron microscopy|room \
temperature|diffraction xrd|electron microscope|results indicate|scanning \
electron|m s|doped zno|microscopy tem|polymer matrix|size distribution|mechanical \
properties|grain size|diameters nm|high spatial|particle size|high resolution|ni \
al|diameter nm|range nm|high field|high strength|c c)'
    >>> trigrams.pattern
    u'(differential scanning calorimetry|face centered cubic|ray microanalysis analytical|\
physical vapor deposition|transmission electron microscopy|x ray diffraction|microanalysis \
analytical electron|chemical vapor deposition|high aspect ratio|analytical electron \
microscope|ray diffraction xrd|x ray microanalysis|high spatial resolution|high \
field magnets|atomic force microscopy|electron microscopy tem|narrow size distribution\
|scanning electron microscopy|building high field|silicon oxide nanowires|particle size \
nm)'
    """

    from nltk.collocations import TrigramCollocationFinder
    from nltk.metrics import BigramAssocMeasures, TrigramAssocMeasures

    # generator of documents, turn each element to its list of words
    documents = (tokenize_simple(text, min_length=min_length, stopwords=stopwords)
                 for text in collection.get_generator_without_id())
    # generator, concatenate (chain) all words into a single sequence, lazily
    words = itertools.chain.from_iterable(documents)
    tcf = TrigramCollocationFinder.from_words(iter(words))

    tcf.apply_freq_filter(min_trigram_freq)
    trigrams = [' '.join(w) for w in tcf.nbest(TrigramAssocMeasures.chi_sq, top_n)]
    logging.info("%i trigrams found: %s..." % (len(trigrams), trigrams[:20]))

    bcf = tcf.bigram_finder()
    bcf.apply_freq_filter(min_bigram_freq)
    bigrams = [' '.join(w) for w in bcf.nbest(BigramAssocMeasures.pmi, top_n)]
    logging.info("%i bigrams found: %s..." % (len(bigrams), bigrams[:20]))

    bigrams_patterns = re.compile('(%s)' % '|'.join(bigrams), re.UNICODE)
    trigrams_patterns = re.compile('(%s)' % '|'.join(trigrams), re.UNICODE)

    return bigrams_patterns, trigrams_patterns


def tokenize_collocation(text, patterns, min_length=1, stopwords=None):
    """A text tokenizer that includes collocations(bigrams and trigrams).

    A collocation is sequence of words or terms that co-occur more often
    than would be expected by chance.  This function breaks a raw document
    up into tokens based on a pre-established collection of bigrams and
    trigrams.  This collection is derived from a body of many documents, and
    must be obtained in a prior step using the collect_bigrams_and_trigrams
    function.

    Uses nltk.collocations.TrigramCollocationFinder to
    find trigrams and bigrams.

    Parameters
    ----------
    text : str
        A single document's text to be tokenized
    patterns: tuple of compiled regex object to find n-grams
        Obtained from collect_bigrams_and_trigrams function
    min_length : int
        Minimum length of any single word
    stopwords : None or iterable of str
        Collection of words to ignore as tokens

    Examples
    --------
    >>> from topik.readers import read_input
    >>> id_documents = read_input('{}/test_data_json_stream.json'.format(test_data_path), content_field="abstract")
    >>> patterns = collect_bigrams_and_trigrams(id_documents, min_bigram_freq=2, min_trigram_freq=2)
    >>> id, doc_text = next(iter(id_documents))
    >>> tokenized_text = tokenize_collocation(doc_text, patterns)
    >>> tokenized_text
    [u'transition_metal', u'oxides', u'considered', u'generation', \
u'materials', u'field', u'electronics', u'advanced', u'catalysts', \
u'tantalum', u'v_oxide', u'reports', u'synthesis_material', \
u'nanometer_size', u'unusual', u'properties', u'work_present', \
u'synthesis', u'ta', u'o', u'nanorods', u'sol', u'gel', u'method', \
u'dna', u'structure', u'directing', u'agent', u'size', u'nanorods', \
u'order', u'nm_diameter', u'microns', u'length', u'easy', u'method', \
u'useful', u'preparation', u'nanomaterials', u'electronics', u'biomedical', \
u'applications', u'catalysts']
    """
    text = ' '.join(tokenize_simple(text, min_length=min_length, stopwords=stopwords))
    for pattern in patterns:
        text = re.sub(pattern, lambda match: match.group(0).replace(' ', '_'), text)
    return text.split()


def collect_entities(collection, freq_min=2, freq_max=10000):
    """Return noun phrases from collection of documents.

    Parameters
    ----------
    collection: Corpus-base derived object or iterable collection of raw text
    freq_min: int
        Minimum frequency of a noun phrase occurrences in order to retrieve it. Default is 2.
    freq_max: int
        Maximum frequency of a noun phrase occurrences in order to retrieve it. Default is 10000.

    """

    from textblob import TextBlob

    np_counts_total = {}
    docs_examined = 0
    for doc in collection.get_generator_without_id():
        if docs_examined > 0 and docs_examined % 1000 == 0:
            sorted_phrases = sorted(np_counts_total.items(),
                                    key=lambda item: -item[1])
            np_counts_total = dict(sorted_phrases)
            logging.info("at document #%i, considering %i phrases: %s..." %
                         (docs_examined, len(np_counts_total), sorted_phrases[0]))

        for np in TextBlob(doc).noun_phrases:
            np_counts_total[np] = np_counts_total.get(np, 0) + 1
        docs_examined += 1

    # Remove noun phrases in the list that have higher frequencies than 'freq_max' or lower frequencies than 'freq_min'
    np_counts = {}
    for np, count in np_counts_total.items():
        if freq_max >= count >= freq_min:
            np_counts[np] = count

    return set(np_counts)


def tokenize_entities(text, entities, min_length=1, stopwords=None):
    """A tokenizer that extracts noun phrases from text.

    Requires that you first establish entities using the collect_entities function

    Parameters
    ----------
    text : str
        A single document's text to be tokenized
    entities : iterable of str
        Collection of noun phrases, obtained from collect_entities function
    min_length : int
        Minimum length of any single word
    stopwords : None or iterable of str
        Collection of words to ignore as tokens

    Examples
    --------
    >>> from topik.readers import read_input
    >>> id_documents = read_input('{}/test_data_json_stream.json'.format(test_data_path), "abstract")
    >>> entities = collect_entities(id_documents)
    >>> len(entities)
    220
    >>> i = iter(id_documents)
    >>> _, doc_text = next(i)
    >>> doc_text
    u'Transition metal oxides are being considered as the next generation \
materials in field such as electronics and advanced catalysts; between\
 them is Tantalum (V) Oxide; however, there are few reports for the \
synthesis of this material at the nanometer size which could have \
unusual properties. Hence, in this work we present the synthesis of \
Ta2O5 nanorods by sol gel method using DNA as structure directing \
agent, the size of the nanorods was of the order of 40 to 100 nm in \
diameter and several microns in length; this easy method can be useful\
 in the preparation of nanomaterials for electronics, biomedical \
applications as well as catalysts.'
    >>> tokenized_text = tokenize_entities(doc_text, entities)
    >>> tokenized_text
    [u'transition']

    """
    from textblob import TextBlob
    result = []
    for np in TextBlob(text).noun_phrases:
        if np in entities:
            # filter out stop words
            tmp = "_".join(tokenize_simple(np, min_length=min_length, stopwords=stopwords))
            # if we end up with nothing, don't append an empty string
            if tmp:
                result.append(tmp)
    return result



def tokenize_mixed(text, entities, min_length=1, stopwords=None):
    """A text tokenizer that retrieves entities ('noun phrases') first and simple words for the rest of the text.

    Parameters
    ----------
    text : str
        A single document's text to be tokenized
    entities : iterable of str
        Collection of noun phrases, obtained from collect_entities function
    min_length : int
        Minimum length of any single word
    stopwords: None or iterable of str
        Collection of words to ignore as tokens

    Examples
    --------
    >>> from topik.readers import read_input
    >>> raw_data = read_input('{}/test_data_json_stream.json'.format(test_data_path), content_field="abstract")
    >>> entities = collect_entities(raw_data)
    >>> id, text = next(iter(raw_data))
    >>> tokenized_text = tokenize_mixed(text, entities, min_length=3)
    >>> tokenized_text
    [u'transition', u'metal', u'oxides', u'generation', u'materials', u'tantalum', \
u'oxide', u'nanometer', u'size', u'unusual', u'properties', u'sol', u'gel', \
u'method', u'dna', u'easy', u'method', u'biomedical', u'applications']

    """
    from textblob import TextBlob
    result = []
    for np in TextBlob(text).noun_phrases:
        if ' ' in np and np not in entities:
            # break apart the noun phrase; it does not occur often enough in the collection of text to be considered.
            result.extend(tokenize_simple(np, min_length=min_length, stopwords=stopwords))
        else:
            # filter out stop words
            tmp = "_".join(tokenize_simple(np, min_length=min_length, stopwords=stopwords))
            # if we end up with nothing, don't append an empty string
            if tmp:
                result.append(tmp)
    return result


# Add additional methods here as necessary to expose them to outside consumers.
tokenizer_methods = {"simple": tokenize_simple,
                     "collocation": tokenize_collocation,
                     "entities": tokenize_entities,
                     "mixed": tokenize_mixed
                     }



#!/usr/bin/env python

from os.path import exists
from setuptools import setup, find_packages
import versioneer

setup(name='topik',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='A Topic Modeling toolkit',
      url='http://github.com/ContinuumIO/topik/',
      author='Topik development team',
      author_email='msarahan@continuum.io',
      license='BSD',
      keywords='topic modeling lda nltk gensim pattern',
      packages=find_packages(),
      package_data={'topik': ['R/runLDAvis.R']},
      install_requires=list(open('requirements.txt').read().strip().split('\n')),
      entry_points= {
          'console_scripts': ['topik = topik.cli:run']
      },
      long_description=(open('README.md').read() if exists('README.md')
                        else ''), zip_safe=False)



from abc import ABCMeta, abstractmethod
from collections import Counter
import logging

import pandas as pd
from six import with_metaclass

# doctest-only imports
from topik.readers import read_input
from topik.tests import test_data_path
from topik.intermediaries.persistence import Persistor

registered_models = {}

def register_model(cls):
    """Decorator function to register new model with global registry of models"""
    global registered_models
    if cls.__name__ not in registered_models:
        registered_models[cls.__name__] = cls
    return cls


class TopicModelBase(with_metaclass(ABCMeta)):
    """Abstract base class for topic models.

    Ensures consistent interface across models, for base result display capabilities.

    Attributes
    ----------
    _corpus : topik.intermediaries.digested_document_collection.DigestedDocumentCollection-derived object
        The input data for this model
    _persistor : topik.intermediaries.persistor.Persistor object
        The object responsible for persisting the state of this model to disk.  Persistor saves metadata
        that instructs load_model how to load the actual data.
    """
    _corpus = None

    @abstractmethod
    def get_top_words(self, topn):
        """Abstract method.  Implementations should collect top n words per topic, translate indices/ids to words.

        Returns
        -------
        list of lists of tuples:
            * outer list: topics
            * inner lists: length topn collection of (weight, word) tuples
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, filename, saved_data):
        """Abstract method.  Persist the model metadata and data to disk.

        Implementations should both save their important data do disk with some known keyword
        (perhaps as filename or server address details), and pass a dictionary to saved_data.
        The contents of this dictionary will be passed to the class' constructor as **kwargs.

        Be sure to either call super(YourClass, self).save(filename, saved_data) or otherwise
        duplicate the base level of functionality here.

        Parameters
        ----------
        filename : str
            The filename of the JSON file to be saved, containing model and corpus metadata
            that allow for reconstruction
        saved_data : dict
            Dictionary of metadata that will be fed to class __init__ method at load time.
            This should include such things as number of topics modeled, binary filenames,
            and any other relevant model parameters to recreate your current model.
        """
        self._persistor.store_model(self.get_model_name_with_parameters(),
                                   {"class": self.__class__.__name__,
                                    "saved_data": saved_data})
        self._corpus.save(filename)

    @abstractmethod
    def get_model_name_with_parameters(self):
        """Abstract method.  Primarily internal function, used to name configurations in persisted metadata for later retrieval."""
        raise NotImplementedError

    def _get_term_data(self):
        vocab = self._get_vocab()
        tf = self._get_term_frequency()
        ttd = self._get_topic_term_dists()
        term_data_df = ttd
        term_data_df['term_frequency'] = tf
        term_data_df['term'] = vocab
        return term_data_df

    def _get_vocab(self):
        return pd.Series(dict(self._corpus._dict.items()))

    def _get_term_frequency(self):
        tf = Counter()
        [tf.update(dict(doc)) for doc in self._corpus]
        # TODO update term documents in intermediate store
        return pd.Series(dict(tf))

    def _get_doc_data(self):
        doc_data_df = self._get_doc_topic_dists()
        doc_data_df['doc_length'] = self._get_doc_lengths()
        return doc_data_df

    def _get_doc_lengths(self):
        id_index, doc_lengths = zip(*[(id, len(doc)) for id, doc in list(
                                                        self._corpus._corpus)])
        return pd.Series(doc_lengths, index=id_index)

    @abstractmethod
    def _get_topic_term_dists(self):
        raise NotImplementedError

    @abstractmethod
    def _get_doc_topic_dists(self):
        raise NotImplementedError

    def to_py_lda_vis(self):
        doc_data_df = self._get_doc_data()
        term_data_df = self._get_term_data()

        model_lda_vis_data = {  'vocab': term_data_df['term'],
                                'term_frequency': term_data_df['term_frequency'],
                                'topic_term_dists': term_data_df.iloc[:,:-2].T,
                                'doc_topic_dists': doc_data_df.iloc[:,:-1],
                                'doc_lengths': doc_data_df['doc_length']}
        return model_lda_vis_data

    def termite_data(self, topn_words=15):
        """Generate the pandas dataframe input for the termite plot.

        Parameters
        ----------
        topn_words : int
            number of words to include from each topic

        Examples
        --------
        >>> raw_data = read_input('{}/test_data_json_stream.json'.format(test_data_path), "abstract")
        >>> processed_data = raw_data.tokenize()  # tokenize returns a DigestedDocumentCollection
        >>> # must set seed so that we get same topics each run
        >>> import random
        >>> import numpy
        >>> random.seed(42)
        >>> numpy.random.seed(42)
        >>> model = registered_models["LDA"](processed_data, ntopics=3)
        >>> model.termite_data(5)
            topic    weight         word
        0       0  0.005337           nm
        1       0  0.005193         high
        2       0  0.004622        films
        3       0  0.004457       matrix
        4       0  0.004194     electron
        5       1  0.005109   properties
        6       1  0.004654         size
        7       1  0.004539  temperature
        8       1  0.004499           nm
        9       1  0.004248   mechanical
        10      2  0.007994         high
        11      2  0.006458           nm
        12      2  0.005717         size
        13      2  0.005399    materials
        14      2  0.004734        phase

        """
        from itertools import chain
        return pd.DataFrame(list(chain.from_iterable([{"topic": topic_id, "weight": weight, "word": word}
                                                      for (weight, word) in topic]
                                                     for topic_id, topic in enumerate(self.get_top_words(topn_words)))))

    @property
    def _persistor(self):
        return self._corpus.persistor


def load_model(filename, model_name):
    """Loads a JSON file containing instructions on how to load model data.

    Returns
    -------
    TopicModelBase-derived object
    """
    p = Persistor(filename)
    if model_name in p.list_available_models():
        data_dict = p.get_model_details(model_name)
        model = registered_models[data_dict['class']](**data_dict["saved_data"])
    else:
        raise NameError("Model name {} has not yet been created.".format(model_name))
    return model
